#include "secure_malloc.h"
#include <sys/mman.h>
#include <unistd.h>
#include <stdalign.h>
#include <stddef.h>

#define ALIGNMENT alignof(max_align_t)

static size_t pagesize = 0;

void* secure_malloc(size_t nbytes) {
   if (!pagesize) pagesize = getpagesize();
   unsigned int nof_pages = (nbytes + ALIGNMENT + pagesize - 1) / pagesize;
   nof_pages += 2; // protecting zone
   void* p = mmap(0, nof_pages * pagesize, PROT_NONE,
      MAP_PRIVATE|MAP_ANON, -1, 0);
   if (p == (void*) MAP_FAILED) return 0;
   unsigned int* block = (unsigned int*)( (char*) p + pagesize );
   if (mprotect((void*)block, (nof_pages-2) * pagesize,
	 PROT_READ|PROT_WRITE) < 0) {
      munmap(p, nof_pages * pagesize);
      return 0;
   }
   *block = nof_pages;
   return (void*) ((char*) block + ALIGNMENT);
}

void secure_free(void* ptr) {
   unsigned int* block = (unsigned int*)((char*) ptr - ALIGNMENT);
   unsigned int nof_pages = *block;
   void* p = (void*) ((char*) block - pagesize);
   munmap(p, nof_pages * pagesize);
}
